/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// put constants into a namespace so we don't pollute the global namespace
function DictState () {
    
    // class contants
    this.Start                = 0;
    this.WantClientAck        = 1;
    this.WantDefinition       = 2;
    this.WantDefinitionsDone  = 3;
    this.WantStrategiesInfo   = 4;
    this.WantStrategies       = 5;
    this.WantMatchesInfo      = 6;
    this.WantMatches          = 7;
    this.Done                 = 8;
    this.Error                = -1;
}

const DICTSTATE = new DictState();



function Observer(connection, word) {
    this._connection = connection;
    this._word       = word;

    this._bufferedData = '';
    this._state = DICTSTATE.Start;

    this._definitions  = {};

    this._createDate = new Date();
}

// Creates a new window and sets the document to be the given HTML.
// Resizes the window to fit the content.
Observer.prototype.createWindow = function (htmlDocument)
{

    const features = [
        'status=no', 
        'toolbar=no',
        'menubar=no',
        'location=no',
        'width=450',
        'height=100'
    ];

    var popupWindow = window.open(
        '',
        '',
        features.join(','));
    
    popupWindow.document.clear();
    popupWindow.document.write(htmlDocument);
    popupWindow.document.close();
    
    // HACK: Figure out how to size the window
    //       to fit the content
    popupWindow.sizeToContent();
    popupWindow.innerHeight += 70;

    // If it's too large to display without scrollbars, shorten it
    if (600 < popupWindow.innerHeight) {
        // limit the height
        popupWindow.innerHeight = 600;
        
        // make wider to account for the width of the 
        // vertical scrollbar.  If we don't do this, a 
        // horizontal scrollbar will appear.
        popupWindow.innerWidth += 20;  
    }

    return popupWindow;
}



function utf8ToCharacterEntitiesInternal(utf8String) {

    var encodedString = '';
    var octetsRemainingInCharacterCode = 0;
    var characterCode = 0;

    for (var i = 0; i < utf8String.length; i++) {

        var inputCharCode = utf8String.charCodeAt(i);

        if ((inputCharCode & 0x80) == 0x00) {
            // this can be encoded with ASCII-7
            encodedString += utf8String[i];

            // reset the multi-octet stream accumulator.
            // if it was not zero, then the input was bad.
            octetsRemainingInCharacterCode = 0;
        }
        else if ((inputCharCode & 0xC0) == 0x80) {
            // 10xxxxxx

            // this is a continuation of a multi-octet sequence
            characterCode = (characterCode << 6) | (inputCharCode & 0x3F);

            octetsRemainingInCharacterCode--;
            if (octetsRemainingInCharacterCode == 0) {
                // this is the final octet in the mutli-octet code
                encodedString += "&#" + characterCode + ";";
            }
            else if (octetsRemainingInCharacterCode < 0) {
                // bad input
                octetsRemainingInCharacterCode = 0;
            }
        }
        else if ((inputCharCode & 0xE0) == 0xC0) {
            // 110xxxxx 10xxxxxx
            characterCode = inputCharCode & 0x1F;
            octetsRemainingInCharacterCode = 1;
        }
        else if ((inputCharCode & 0xF0) == 0xE0) {
            // 1110xxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x0F;
            octetsRemainingInCharacterCode = 2;
        }
        else if ((inputCharCode & 0xF8) == 0xF0) {
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x07;
            octetsRemainingInCharacterCode = 3;
        }
        else if ((inputCharCode & 0xFC) == 0xF8) {
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x03;
            octetsRemainingInCharacterCode = 4;
        }
        else if ((inputCharCode & 0xFE) == 0xFC) {
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x01;
            octetsRemainingInCharacterCode = 5;
        }
        else {
            // Illegal UTF-8 code.  Drop it.
        }
    }

    return encodedString;
}

//
// utf8ToCharacterEntities() converts an octet stream that is 
// encoded in UTF-8 into an ASCII-7 encoding of the same 
// stream such that all character > 127 are encoded using
// HTML character entities &#xxx;.
//
// This is necessary when using international DICT servers,
// because I cannot figure out another way to tell Mozilla
// to interpret the string as UTF-8.
//
//   UCS-4 range (hex.)    UTF-8 octet sequence (binary)
//   0000 0000-0000 007F   0xxxxxxx
//   0000 0080-0000 07FF   110xxxxx 10xxxxxx
//   0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
//   0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//
function utf8ToCharacterEntities(utf8String) {

    // Before we do any convertion, let's see if there are
    // any character codes outside of the ASCII-7 range.
    // If not, then the string doesn't need to be converted.

    if (!utf8String.match(/[\x80-\xFF]/)) {
        return utf8String;
    }

    // Note that we could just call utf8ToCharacterEntitiesInternal()
    // on the entire string, but this routine is very slow.
    // Instead, it is faster to use a regular expression
    // to separate the octet-sequences that must be replaced
    // while preserving long runs of characters that don't to escaping.

    // This makes the routine run in O(n^2) of the characters that
    // must be replaced, instead of O(n^2) of all characters.
    // When the text is mostly ASCII-7, this is a significant savings.

    var encodedString = '';
    
    var matchExp = /([\x00-\x7F]*)([\x80-\xFF]+)([\x00-\x7F]*)/g;
    while ((result = matchExp.exec(utf8String)) != null) {

        var singleOctetCodesBefore = result[1];
        var multiOctetCodes        = result[2];
        var singleOctetCodesAfter  = result[3];
        
        encodedString += 
            singleOctetCodesBefore + 
            utf8ToCharacterEntitiesInternal(multiOctetCodes) +
            singleOctetCodesAfter;
    }

    return encodedString;
}


// wraps some HTML inside HTML/BODY tag and adds a title
Observer.prototype.createHtmlDocument = function (title, body)
{
    var htmlDocument;
    
    htmlDocument = 
        '<HTML>\n'    +
        '  <HEAD>\n'  + 
        '      <TITLE>' + utf8ToCharacterEntities(title) + '</TITLE>\n' +
        '  </HEAD>\n' +
        '  <BODY>\n'  +
        utf8ToCharacterEntities(body) +
        '  </BODY>\n' +
        '</HTML>\n';

    return htmlDocument;
}

// Reports that there was an protocol error.
// This should only happen if the server doesn't obey the protocol
// (or if there are bugs in the client state machine)
Observer.prototype.reportStateMachineError = function ()
{
    throw 
        'Unknown problem encountered when defining "' + this._word + '"\n' +
        'If problem presists, please file a bug report on the Dict project\n';
}

// Pops a DICT response line from the _bufferedInput member variable.
// On success, returns true and sets this._responseLine, this._responseCode, 
// and this._responseLineBody.
// Return false, if there wasn't a complete response line.
Observer.prototype.popResponseLine = function ()
{
    var responseLinePattern = /^(\d+)\s*(.*)\n/;
    
    var result = this._bufferedData.match(responseLinePattern);
    if (result == null) {
        return false;
    }

    this._responseLine     = result[0];
    this._responseCode     = result[1];
    this._responseLineBody = result[2];
    
    // remove the response line from the buffered input
    var lineEndIndex   = this._bufferedData.indexOf('\n');
    this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);
    return true;
}

// Performs a graceful shutdown
Observer.prototype.gracefulShutdown = function ()
{
    this._connection.sendData('QUIT\r\n');
    this._bufferedData = '';
    this._state = DICTSTATE.Done;
    
    this._connection.disconnect();
}


// Parses and returns all buffered input up to (but not including) 
// the sequence CRLF . CRLF.  The multiline input is then removed 
// from the buffered input.
// Returns null if the multiline input is not complete.
//
// The DICT protocol uses multiline input of this form to return 
// strategies, matches, and definitions. 
Observer.prototype.popMultilineInput = function () 
{
    // NOTE: We should like for \r\n.\r\n, but all "\r" are removed
    //       from the input stream to make regexp parsing easier.
    const sentinal = "\n.\n";

    // Check to see if we have the entire list
    var endIndex = this._bufferedData.indexOf(sentinal);
    if (endIndex == -1) {
        return null;  // the multiline input isn't complete
    }
    
    // pop off and return the multiline input
    var multilineData = this._bufferedData.slice(0, endIndex);
    
    // remove the multiline input and sentinal sequence from the buffered input
    this._bufferedData = this._bufferedData.slice(endIndex + sentinal.length);

    return multilineData;
}

Observer.prototype.onStreamDataAvailable = function (
    request, 
    inStream, 
    offset, 
    count
) 
{
    try {
        // read the data from the socket into our buffer
        var newData = this._connection.readData(0, count);

        // strip out carriage returns, since they complicate regexp usage
        this._bufferedData += newData.replace(/\r/g, '');

        // do {
        //   {handle data}
        // } while( {there's enough data to handle} );
        do {
            var previousLength = this._bufferedData.length;

            switch (this._state) {

            case DICTSTATE.Start:
                // get the 220 "hello" response that is sent when we connect

                if (this.popResponseLine()) {
                    
                    if (this._responseCode == 220) {
                        // 220 <text> <capabilities> <msg-id> CRLF
                        // (this is the session init sequence)
                    
                        this._state = DICTSTATE.WantClientAck;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;


            case DICTSTATE.WantClientAck:
                // get the 250 ok resonse for our "CLIENT" command

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        this._state = DICTSTATE.WantDefinitionsDone;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinitionsDone:

                if (this.popResponseLine()) {
                    if (this._responseCode == 150) {
                        // 150: n definitions returned
                    }
                    else if (this._responseCode == 151) {
                        // 151: <word> <database> <name>
                        // The definition follows this line
                        this._state = DICTSTATE.WantDefinition;
                    }
                    else if (this._responseCode == 250) {

                        // We have all definitions

                        // Format the definitions into HTML
                        var definitionDocumentBody = '<TABLE width="450"><TBODY><TR><TD>';

                        for (source in this._definitions) {

                            definitionDocumentBody += '<B>';
                            definitionDocumentBody += source;
                            definitionDocumentBody += '</B><BR>\n';
                        
                            var i;
                            for (i = 0; i < this._definitions[source].length; i++) {

                                var deftext = this._definitions[source][i].text;
                                
                                // Escape HTML entities
                                deftext = deftext.replace(/&/g,   '&amp;');
                                deftext = deftext.replace(/</g,   '&lt;');
                            
                                // replace newlines with <BR> and whitespace with
                                // non-breaking spaces to preserve formatting.
                                // This is better than using <PRE> because it
                                // allows us to use other mark-up tags.
                                deftext = deftext.replace(/\n/g, '<BR>');
                                
                                // Words that are cross-referenced by WordNet show up 
                                // in braces. e.g., "{word}" and "{word(P)}"
                                deftext = deftext.replace(/{(.*?)(\(.*\))?}/g, '$1');

                                // add newlines after <BR> for readability
                                deftext = deftext.replace(/<BR>/g, '\n');

                                definitionDocumentBody += '<PRE>'+deftext+'\n</PRE>\n';
                            }
                        }

                        definitionDocumentBody += '</TD></TR></TBODY></TABLE>\n';

                        var definitionDocument = this.createHtmlDocument(
                            this._word,
                            definitionDocumentBody);

                        this.createWindow(definitionDocument);

                        this.gracefulShutdown();
                    }
                    else if (this._responseCode == 550) {
                        // 550 invalid database

                        // TODO: Handle this gracefully
                        this.reportStateMachineError();
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match

                        // Present the user with a nearest match 
                        
                        // The "*" database means search through all databases.
                        //
                        // The "." strategy means "use whatever is best for
                        // for interactive spell-checking", according to 
                        // RFC 2229, section 3.3.1.
                        //
                        this._connection.sendData('MATCH * . "' + this._word + '"\r\n');
                        this._state = DICTSTATE.WantMatchesInfo;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinition:
                
                // 151 <word> <database> CRLF
                // <definition line 1>   CRLF
                // <definition line 2>   CRLF
                // . . . 
                // <definition line n>   CRLF
                // . CRLF
                
                // determine if we have the entire definition
                var definitionText = this.popMultilineInput();
                if (definitionText == null) {
                    break; // wait for the rest of the definition
                }

                var definition = new Object;

                // 151 <word> <database> CRLF
                var result = this._responseLineBody.match(/^(\S+)\s+\S+\s+(.*)$/);
                if (result != null) {
                    definition.word   = result[1];
                    definition.source = result[2];
                }

                // Store the definition
                definition.text = definitionText;

                // Add this definition to the set of definitions.
                // this._definitions maps the source name to an array of 
                // definition objects.
                if (this._definitions[definition.source] != null) {
                    this._definitions[definition.source].push(definition);
                }
                else {
                    this._definitions[definition.source] = new Array ();
                    this._definitions[definition.source].push(definition);
                }

                this._state = DICTSTATE.WantDefinitionsDone;
                break;

            case DICTSTATE.WantStrategiesInfo:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   this state code is correct, so I have kept it intact.

                if (this.popResponseLine()) {
                    if (this._responseCode == 111) {
                        // 111 n stragegies available
                        this._state = DICTSTATE.WantStrategies;
                    }
                    else if (this._responseCode == 555) {
                        // 555 No strategies available
                        this.gracefulShutdown();
                        alert('No definitions found for "' + this._word + '"');
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantStrategies:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   the state code is correct, so I have kept it intact.

                // <strategy 1> <description 1> CRLF
                // <strategy 2> <description 2> CRLF
                // . . . 
                // <strategy n> <description n> CRLF
                // . CRLF

                // determine if we have the entire strategy list
                var strategies = this.popMultilineInput();
                if (strategies == null) {
                    break; // wait for the rest of the strategies
                }


                // search for the "Levenshtein distance one"
                var haveLevStrategy = false;
                // <strategy> <description> CRLF
                var matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(strategies)) != null) {
                    var strategy    = result[1];
                    var description = result[2];

                    if (strategy == 'lev') {
                        haveLevStrategy = true;
                    }
                }

                if (haveLevStrategy) {
                    // Use the "lev" strategy to find nearest matches
                    this._connection.sendData('MATCH * lev "' + this._word + '"\r\n');
                    this._state = DICTSTATE.WantMatchesInfo;
                }
                else {
                    // No good matching strategies
                    this.gracefulShutdown();
                    
                    alert('No definitions found for "' + this._word + '"');
                }

                break;

            case DICTSTATE.WantMatchesInfo:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        // 250 ok
                        // This is from the MATCH request.  Ignore it.
                    }
                    else if (this._responseCode == 152) {
                        // 152 n matches found
                        this._state = DICTSTATE.WantMatches;
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match
                        this.gracefulShutdown();
                        alert('No definitions found for "' + this._word + '"');
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantMatches:

                // <database 1> <word 1> CRLF
                // <database 2> <word 2> CRLF
                // . . . 
                // <database n> <word n> CRLF
                // . CRLF
                
                // determine if we have the entire strategy list
                var matches = this.popMultilineInput();
                if (matches == null) {
                    break; // wait for the rest of the matches
                }

                // <database> <word> CRLF
                var words    = new Object;
                var matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(matches)) != null) {
                    var database    = result[1];
                    var word        = result[2];

                    var normalizedWord = word.toLowerCase();

                    words[normalizedWord] = true;
                }

                // REVISIT: There has got to be a more direct way
                //   of counting the number of elements in an associative
                //   array.
                var totalMatches = 0;
                for (word in words) {
                    totalMatches++;
                }

                if (totalMatches == 0) {
                    // There are no matches.

                    // This is probably a bug in either the server or the
                    // client, since we should not have transitioned to
                    // the WantMatches state if there were no matches.
                    this.gracefulShutdown();

                    alert('No definitions found for "' + this._word + '"');
                }
                else if (totalMatches == 1) {
                    // There is exactly one match.
                    // Rather than present just one match, we can save the
                    // user some trouble and just request the definition
                    // of the match.

                    var word;
                    for (word in words) {
                        // Request a defintion for the only match
                        this._connection.sendData('DEFINE * ' + word + '\r\n');
                        this._word  = word;
                        this._state = DICTSTATE.WantClientAck;
                    }
                }
                else {
                    // There is more than one closest match
                    // Display them all and let the user decide which to define
                    var matchesDocumentBody = 
                        '<b>Closest Matches</b><br>\n' + 
                        '<table width="450" border="0"><tr><td>'  +
                        'No definitions for "' + this._word + '" were found, ' +
                        'but definitions for similar-looking words were found.  ' +
                        'To define one of these close matches, highlight ' +
                        'the word below, right-click, and select "Define" ' +
                        'from the pop-up menu.';

                    // add a spacer between the text and list of words
                    matchesDocumentBody += '<tr><td><hr></td></td>\n';

                    var word;
                    for (word in words) {
                        matchesDocumentBody += '<tr><td>' + word + '</td></td>\n';
                    }

                    matchesDocumentBody += '</td></tr></table>\n';

                    var matchesDocument = this.createHtmlDocument(
                        this._word + ' - Closest Matches',
                        matchesDocumentBody);

                    this.createWindow(matchesDocument);
                    
                    this.gracefulShutdown();
                }

                break;

            case DICTSTATE.Done:
                break;

            default:
                reportStateMachineError();
                break;
            }
            
        } while (previousLength != this._bufferedData.length);
    }
    catch (ex) {
        this._connection.disconnect();
        alert(ex);
    }
}

Observer.prototype.onStreamClose = function (
    status
) 
{
    var closeDate = new Date();
    var delta = closeDate.getTime() - this._createDate.getTime();
    //alert('query took ' + delta + ' ms');
}

function DictClient() {
}

DictClient.prototype.defineWord = function(host, port, word)
{
    var connection = new CBSConnection();
    
    connection.connect(host, port, (void 0), true, null);
    
    var observer = new Observer(connection, word);

    // To save time, pipe-line the CLIENT and DEFINE commands,
    // rather than wait for the "220 ok" handshake.
    // The server is required to support command pipe-lining.
    // This makes a measurable difference there is high latency.
    connection.sendData('CLIENT Mozilla Dict plugin\r\n');
    connection.sendData('DEFINE * "' + word + '"\r\n');
    connection.startAsyncRead(observer);
}

